//----------------------------------------------------------------------------
//
// Author              : Jan Horn
// Email               : jhorn@global.co.za
// Website             : http://www.sulaco.co.za
//                     : http://home.global.co.za/~jhorn
// Date                : 8 April 2001
// TMT Pascal version  : Vadim Bodrov (http://www.tmt.com)
// Description         : A unit that used with OpenGL projects to load BMP files
// Usage               : LoadTexture(BMPFilename, TextureName);
//                       eg : LoadTexture('logo.bmp', LogoTex);
//
// Changes             : 28 July    - Faster BGR to RGB routine
//                     : 27 January - Added load from resource
//----------------------------------------------------------------------------
unit BMP2;

interface

uses
  Strings, Windows, OpenGL;

function LoadTexture(Filename: String; var Texture : GLuint) : Boolean;
function LoadResTexture(ResFilename: String; var Texture : GLuint) : Boolean;

implementation

procedure glBindTexture conv arg_stdcall (target: GLenum; texture: GLuint);
  external opengl32dll;
procedure glGenTextures conv arg_stdcall (n: GLsizei; var textures: GLuint);
  external opengl32dll;

{------------------------------------------------------------------}
{  Swap bitmap format from BGR to RGB                              }
{------------------------------------------------------------------}
procedure SwapRGB(data : Pointer; Size : Integer); assembler;
asm
  mov ebx, [data]
  mov ecx, [size]

@@loop :
  mov al,[ebx+0]
  mov ah,[ebx+2]
  mov [ebx+2],al
  mov [ebx+0],ah
  add ebx,3
  dec ecx
  jnz @@loop
end;


{------------------------------------------------------------------}
{  Create the Texture                                              }
{------------------------------------------------------------------}
function CreateTexture(Width, Height, Format : Word; pData : Pointer) : Integer;
var
  Texture : GLuint;
begin
  glGenTextures(1, Texture);
  glBindTexture(GL_TEXTURE_2D, Texture);
  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);  {Texture blends with object background}
//  glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_DECAL);  {Texture does NOT blend with object background}

  { Select a filtering type. BiLinear filtering produces very good results with little performance impact
    GL_NEAREST               - Basic texture (grainy looking texture)
    GL_LINEAR                - BiLinear filtering
    GL_LINEAR_MIPMAP_NEAREST - Basic mipmapped texture
    GL_LINEAR_MIPMAP_LINEAR  - BiLinear Mipmapped texture
  }

  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR); { only first two can be used }
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR); { all of the above can be used }

  if Format = GL_RGBA then
    gluBuild2DMipmaps(GL_TEXTURE_2D, GL_RGBA, Width, Height, GL_RGBA, GL_UNSIGNED_BYTE, pData)
  else
    gluBuild2DMipmaps(GL_TEXTURE_2D, 3, Width, Height, GL_RGB, GL_UNSIGNED_BYTE, pData);
//  glTexImage2D(GL_TEXTURE_2D, 0, 3, Width, Height, 0, GL_RGB, GL_UNSIGNED_BYTE, pData);  // Use when not wanting mipmaps to be built by openGL

  result :=Texture;
end;


{------------------------------------------------------------------}
{  Load BMP textures from file                                     }
{------------------------------------------------------------------}
function LoadTexture(Filename: String; var Texture : GLuint) : Boolean;
var
  FileHeader : TBitmapFileHeader;
  InfoHeader : TBitmapInfoHeader;
  Palette    : Pointer;
  pData      : Pointer;
  BitmapFile    : THandle;
  BitmapLength  : LongWord;
  PaletteLength : LongWord;
  ReadBytes     : LongWord;
  Width, Height : Integer;
  Buff          : array[0..MAX_PATH] of char;
begin
  result :=FALSE;

  BitmapFile := CreateFile(StrPCopy(Buff, Filename), GENERIC_READ, FILE_SHARE_READ, nil, OPEN_EXISTING, 0, 0);
  if (BitmapFile = INVALID_HANDLE_VALUE) then begin
    MessageBox(0, StrPCopy(Buff, 'Error opening ' + Filename), PChar('BMP Unit'), MB_OK);
    Exit;
  end;

  // Get header information
  ReadFile(BitmapFile, FileHeader, SizeOf(FileHeader), ReadBytes, nil);
  ReadFile(BitmapFile, InfoHeader, SizeOf(InfoHeader), ReadBytes, nil);

  // Get palette
  PaletteLength := InfoHeader.biClrUsed;
  GetMem(Palette, PaletteLength);
  ReadFile(BitmapFile, Palette, PaletteLength, ReadBytes, nil);
  if (ReadBytes <> PaletteLength) then begin
    MessageBox(0, PChar('Error reading palette'), PChar('BMP Unit'), MB_OK);
    Exit;
  end;

  Width  := InfoHeader.biWidth;
  Height := InfoHeader.biHeight;
  BitmapLength := InfoHeader.biSizeImage;
  if BitmapLength = 0 then
    BitmapLength := Width * Height * InfoHeader.biBitCount Div 8;

  // Get the actual pixel data
  GetMem(pData, BitmapLength);
  ReadFile(BitmapFile, pData^, BitmapLength, ReadBytes, nil);
  if (ReadBytes <> BitmapLength) then begin
    MessageBox(0, PChar('Error reading bitmap data'), PChar('BMP Unit'), MB_OK);
    Exit;
  end;
  CloseHandle(BitmapFile);

  // Bitmaps are stored BGR and not RGB, so swap the R and B bytes.
  SwapRGB(pData, Width*Height);

  Texture :=CreateTexture(Width, Height, GL_RGB, pData);
  FreeMem(pData, BitmapLength);
  GetMem(Palette, PaletteLength);
  result :=TRUE;
end;


{------------------------------------------------------------------}
{  Load BMP textures from the app resource                         }
{------------------------------------------------------------------}
function LoadResTexture(ResFilename: String; var Texture : GLuint) : Boolean;
var
  InfoHeader : TBitmapInfoHeader;
  Palette    : Pointer;
  pData      : Pointer;
  BitmapLength  : LongWord;
  PaletteLength : LongWord;
  Width, Height : Integer;
  Buff          : array[0..MAX_PATH] of char;
  rsrc          : HRSRC;
  global        : HGLOBAL;
  resptr        : Pointer;
begin
  result := FALSE;

  rsrc := FindResource(hInstance, StrPCopy(Buff, copy(ResFilename, 1, Pos('.', ResFilename)-1)), 'BMP');
  if rsrc = 0 then
  begin
    MessageBox(0, StrPCopy(Buff, 'File not found in resource - ' + ResFilename), PChar('BMP Texture'), MB_OK);
    Exit;
  end;

  global := LoadResource(hInstance, rsrc);
  if global = 0 then
  begin
    MessageBox(0, StrPCopy(Buff, 'Unable to read from resource - ' + ResFilename), PChar('BMP Unit'), MB_OK);
    Exit;
  end;

  resptr := LockResource(global);

  Move(Pointer(resptr + SizeOf(TBitmapFileHeader))^, InfoHeader, SizeOf(TBitmapInfoHeader));
  PaletteLength := InfoHeader.biClrUsed;
  Palette := Pointer(resptr + SizeOf(TBitmapFileHeader) + SizeOf(TBitmapInfoHeader));

  Width  := InfoHeader.biWidth;
  Height := InfoHeader.biHeight;

  BitmapLength := InfoHeader.biSizeImage;
  if BitmapLength = 0 then
    BitmapLength := Width * Height * InfoHeader.biBitCount Div 8;

  pData := Pointer(resptr + SizeOf(TBitmapFileHeader) + SizeOf(TBitmapInfoHeader) + PaletteLength);

  // Bitmaps are stored BGR and not RGB, so swap the R and B bytes.
  SwapRGB(pData, Width*Height);

  Texture :=CreateTexture(Width, Height, GL_RGB, pData);
  result :=TRUE;
end;

end.